from typing import TypeVar, Generic, Callable, List

from treepath.descriptor.document import Document
from treepath.path.typing.json_types import JsonTypes
from treepath.path.utils.function import do_nothing

T = TypeVar('T')


class DocumentList(Generic[T], Document):
    """
    DocumentList decorates a json list.  The DocumentList (un)marshals each entry in the json list to service each of
    the basic list operations.    These operations include len(), iter(), [], []=, del [], in, pop, append.
    """
    __slots__ = "_data", "_to_wrapped_value", "_to_json_value"

    def __init__(self,
                 data: List[JsonTypes],
                 *,
                 to_wrapped_value: Callable[[JsonTypes], T] = do_nothing,
                 to_json_value: Callable[[T], JsonTypes] = do_nothing):
        super().__init__(data)
        self._to_wrapped_value = to_wrapped_value
        self._to_json_value = to_json_value

    def __iter__(self) -> T:
        class _Iterator(Generic[T]):
            __slots__ = "_iterator", "_to_wrapped_value"

            def __init__(self, iterator, to_wrapped_value):
                self._to_wrapped_value = to_wrapped_value
                self._iterator = iterator

            def __next__(self) -> T:
                json_value = next(self._iterator)
                return self._to_wrapped_value(json_value)

        return _Iterator[T](iter(self._data), self._to_wrapped_value)

    def __getitem__(self, index: int) -> T:
        json_value = self.data[index]
        wrapped_value = self._to_wrapped_value(json_value)
        return wrapped_value

    def __setitem__(self, index: int, wrapped_value: T):
        json_value = self._to_json_value(wrapped_value)
        self.data[index] = json_value

    def __len__(self) -> int:
        return len(self.data)

    def __delitem__(self, index):
        del self.data[index]

    def __contains__(self, wrapped_value: T):
        json_value = self._to_json_value(wrapped_value)
        return json_value in self.data

    def append(self, wrapped_value: T):
        json_value = self._to_json_value(wrapped_value)
        self.data.append(json_value)

    def pop(self, index: int) -> T:
        json_value = self.data.pop(index)
        wrapped_value = self._to_wrapped_value(json_value)
        return wrapped_value

    def remove_all(self, is_remove: Callable[[T], bool]):
        def is_not_keep(arg):
            return not is_remove(arg)

        self.keep_all(is_not_keep)

    def keep_all(self, is_keep: Callable[[T], bool]):
        data = self.data
        original_length = len(data)
        last_index = -1
        for last_index, json_value in enumerate(map(self._to_json_value, filter(is_keep, self))):
            data[last_index] = json_value
        for index in range(last_index + 1, original_length):
            data.pop()





class AbstractPathBuilder:
    __slots__ = ()

    def create_path_builder(self, *args, **kwargs):
        raise NotImplementedError  # pragma: no cover

    def tranform_attribute_name(self, name):
        raise NotImplementedError  # pragma: no cover



from treepath.path.typing.json_arg_types import JsonArgTypes


class AbstractDocument:
    @property
    def data(self) -> JsonArgTypes:  # pragma: no cover
        ...

    @data.setter
    def data(self, data: JsonArgTypes):  # pragma: no cover
        ...



from typing import TypeVar, Generic, Callable, Iterable

from treepath.path.typing.json_types import JsonTypes
from treepath.path.utils.function import do_nothing

T = TypeVar('T')


class DocumentIterator(Generic[T]):
    """
    DocumentIterator decorates a json iterator.  The  DocumentIterator marshals each entry in the json iterator
    before return it to the caller.
    """
    __slots__ = "_iterable", "_to_wrapped_value"

    def __init__(self,
                 iterable: Iterable[JsonTypes],
                 *,
                 to_wrapped_value: Callable[[JsonTypes], T] = do_nothing):
        self._iterable = iterable
        self._to_wrapped_value = to_wrapped_value

    def __iter__(self) -> T:
        class _Iterator(Generic[T]):
            __slots__ = "_iterator", "_to_wrapped_value"

            def __init__(self, iterator, to_wrapped_value):
                self._to_wrapped_value = to_wrapped_value
                self._iterator = iterator

            def __next__(self) -> T:
                json_value = next(self._iterator)
                return self._to_wrapped_value(json_value)

        return _Iterator[T](iter(self._iterable), self._to_wrapped_value)



from typing import Union, Callable, Any

from treepath.path.builder.path_builder import PathBuilder
from treepath.path.traverser.match import Match
from treepath.path.traverser.traverser_functions import get, set_
from treepath.path.traverser.traverser_functions import get_match


def pprop(path: PathBuilder,
          get_data: Union[property, Callable[[Any], Union[dict, list, str, int, float, bool, None]]]
          ):
    """
    Deprecated use treepath.attr instead
    """
    return __DeprecatedPathProperty(path, get_data)


def mprop(path: PathBuilder,
          get_data: Union[property, Callable[[Any], Union[dict, list, str, int, float, bool, None]]]
          ):
    """
    Deprecated use treepath.attr instead
    """
    return __DeprecatedMatchPathProperty(path, get_data)


class __DeprecatedPathProperty:
    __slots__ = '_get_data', \
                '_path'

    def __init__(self,
                 path: PathBuilder,
                 get_data: Union[property, Callable[[Any], Union[dict, list, str, int, float, bool, None]]]
                 ):
        self._path = path
        self._set_get_data(get_data)

    def _set_get_data(self, get_data):
        """
        Determine how data, is a callable, property or a descriptor.
        """
        if callable(get_data):
            self._get_data = get_data
        elif isinstance(get_data, property):
            self._get_data = get_data.fget
        elif hasattr(get_data, '__get__'):
            def getter(instance):
                return get_data.__get__(instance, None)

            self._get_data = getter
        else:
            raise AttributeError(f"Cannot invoke f{get_data}")  # pragma: no cover

    def get_value(self, outer_self):
        return get(self._path, self._get_data(outer_self), default=None)

    def get_match(self, outer_self) -> Match:
        return get_match(self._path, self._get_data(outer_self), must_match=False)

    def set_value(self, outer_self, value):
        set_(self._path, value, self._get_data(outer_self), cascade=True)

    def __get__(self, instance, owner) -> Union[dict, list, str, int, float, bool, None]:
        return self.get_value(instance)

    def __set__(self, instance, value: Union[dict, list, str, int, float, bool, None]):
        return self.set_value(instance, value)


class __DeprecatedMatchPathProperty(__DeprecatedPathProperty):

    def __init__(self, path: PathBuilder,
                 get_data: Union[property, Callable[[Any], Union[dict, list, str, int, float, bool, None]]]):
        super().__init__(path, get_data)

    def __get__(self, instance, owner) -> Match:
        return self.get_match(instance)



from treepath.path.traverser.match import Match
from treepath.path.builder.path_builder import PathBuilder
from treepath.path.builder.root_path_builder import RootPathBuilder


def match_to_path(match: Match) -> PathBuilder:
    """
    Converts a match to a Path
    """
    path_as_list = match.path_match_list
    path = RootPathBuilder()
    for match in path_as_list[1:]:
        data_name = match.data_name
        path = path[data_name]
    return path



from treepath.path.traverser.nested_match_traverser import NestedMatchTraverser
from treepath.path.typing.json_types import JsonTypes


class NestedValueTraverser(NestedMatchTraverser):
    __slots__ = ()

    def __next__(self) -> JsonTypes:
        return super().__next__().data



from __future__ import annotations

from typing import Callable, Union, Any, Type, Optional

from treepath.descriptor.descriptor_builder import DescriptorBuilder
from treepath.descriptor.path_descriptor import PathDescriptor
from treepath.descriptor.path_descriptor import T
from treepath.path.builder.path_builder import PathBuilder
from treepath.path.traverser.traverser_functions import get, set_, find, get_match, find_matches, set_match
from treepath.path.typing.json_types import JsonTypes
from treepath.path.utils.function import do_nothing


def attr(expression: Optional[PathBuilder] = None,
         *,
         getter: Union[get, find, get_match, find_matches] = get,
         setter: Union[set_, set_match] = set_,
         to_wrapped_value: Callable[[JsonTypes], Any] = do_nothing,
         to_json_value: Callable[[Any], JsonTypes] = do_nothing
         ) -> PathDescriptor[T]:
    """
    Constructs a descriptor to be attached to a subclass of a Document.

    class X(Document):
        a = attr()

    x = X({"a": 1})
    assert x.a == 1

    @param expression: The path expression that define the search criteria. The expression is optional.  The default
      value is the attribute name the descriptor is being assigned too.
    @param getter: The treepath traverser function used to fetch the value from the Document.
    @param setter: The treepath traverser function used to assign the value to the Document.
    @param to_wrapped_value: The treepath function used to unmarshal the value fetch using the getter
    @param to_json_value: The treepath function used to marshal the value to be assigned using the setter.

    @return PathDescriptor:  Returns a path descriptor configured as specified by the parameters.
    """
    return PathDescriptor(
        expression=expression,
        getter=getter,
        setter=setter,
        to_wrapped_value=to_wrapped_value,
        to_json_value=to_json_value,
    )


def attr_typed(type_: Type[T],
               expression: Optional[PathBuilder] = None,
               *,
               getter: Union[get, find, get_match, find_matches] = get,
               setter: Union[set_, set_match] = set_,
               to_wrapped_value: Callable[[JsonTypes], Any] = do_nothing,
               to_json_value: Callable[[Any], JsonTypes] = do_nothing
               ) -> PathDescriptor[T]:
    """
    Constructs Typed descriptor to be attached to a subclass of a treepath Document.

    class X(Document):
        a = attr()

    class Y(Document):
        b = attr_typed(X)

    y = Y({"b": {"a": 1}})
    assert y.b.a == 1

    @param type_: The attribute type.  The type can be either subclass of treepath Document or a custom type.  The
    to_wrapped_value and to_json_value must be defined for custom type.
    @param expression: The path expression that define the search criteria. The expression is optional.  The default
      value is the attribute name the descriptor is being assigned too.
    @param getter: The treepath traverser function used to fetch the value from the Document.
    @param setter: The treepath traverser function used to assign the value to the Document.
    @param to_wrapped_value: The treepath function used to unmarshal the value fetch using the getter.  The argument
      must be specified for custom types.
    @param to_json_value: The treepath function used to marshal the value to be assigned using the setter.  The
      argument must be specified for custom types.
    @return PathDescriptor:  Returns a path descriptor configured as specified by the parameters.
    """
    descriptor_builder = DescriptorBuilder(
        type_=type_,
        expression=expression,
        getter=getter,
        setter=setter,
        to_wrapped_value=to_wrapped_value,
        to_json_value=to_json_value,
    )
    return descriptor_builder.build_single()


def attr_iter_typed(type_: Type[T],
                    expression: Optional[PathBuilder] = None,
                    *,
                    getter: Union[find, find_matches] = find,
                    to_wrapped_value: Callable[[JsonTypes], Any] = do_nothing,
                    ) -> PathDescriptor[T]:
    """
    Constructs Typed iterator descriptor to be attached to a subclass of a treepath Document.

    class X(Document):
        a = attr()

    class Y(Document):
        b = attr_iter_typed(X, path.b[wc])

    y = Y({"b": [{"a": 1}, {"a": 2}]})
    x = iter(y.b)
    assert next(x).a == 1
    assert next(x).a == 2

    @param type_: The attribute type.  The type can be either subclass of treepath Document or a custom type.  The
    to_wrapped_value and to_json_value must be defined for custom type.
    @param expression: The path expression that define the search criteria. The expression is optional.  The default
      value is the attribute name the descriptor is being assigned too.
    @param getter: The treepath traverser function used to fetch the value from the Document.
    @param to_wrapped_value: The treepath function used to unmarshal the value fetch using the getter.  The argument
      must be specified for custom types.
    @return PathDescriptor:  Returns a path descriptor configured as specified by the parameters.
    """
    descriptor_builder = DescriptorBuilder(
        type_=type_,
        expression=expression,
        getter=getter,
        to_wrapped_value=to_wrapped_value,
    )
    return descriptor_builder.build_iterator()


def attr_list_typed(type_: Type[T],
                    expression: Optional[PathBuilder] = None,
                    *,
                    getter: Union[get, get_match] = get,
                    to_wrapped_value: Callable[[JsonTypes], Any] = do_nothing,
                    to_json_value: Callable[[Any], JsonTypes] = do_nothing
                    ) -> PathDescriptor[T]:
    """
    Constructs Typed list descriptor to be attached to a subclass of a treepath Document.

    class X(Document):
        a = attr()

    class Y(Document):
        b = attr_list_typed(X, path.b)

    y = Y({"b": [{"a": 1}, {"a": 2}]})
    assert y.b[0].a == 1
    assert y.b[1].a == 2

    @param type_: The attribute type.  The type can be either subclass of treepath Document or a custom type.  The
    to_wrapped_value and to_json_value must be defined for custom type.
    @param expression: The path expression that define the search criteria. The expression is optional.  The default
      value is the attribute name the descriptor is being assigned too.
    @param getter: The treepath traverser function used to fetch the value from the Document.
    @param to_wrapped_value: The treepath function used to unmarshal the value fetch using the getter.  The argument
      must be specified for custom types.
    @param to_json_value: The treepath function used to marshal the value to be assigned using the setter.  The
      argument must be specified for custom types.
    @return PathDescriptor:  Returns a path descriptor configured as specified by the parameters.
    """
    descriptor_builder = DescriptorBuilder(
        type_=type_,
        expression=expression,
        getter=getter,
        to_wrapped_value=to_wrapped_value,
        to_json_value=to_json_value,
        is_for_json_list=True
    )
    return descriptor_builder.build_list()



from typing import Union

from treepath.path.exceptions.set_error import SetError
from treepath.path.traverser.list_match import ListMatch
from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.typing.json_types import JsonTypes
from treepath.path.utils.function import enumerate_slice
from treepath.path.vertex.vertex import Vertex


class _ListVertex(Vertex):
    __slots__ = ()

    def __init__(self, parent, name):
        super().__init__(parent, name)

    @property
    def path_segment(self):
        return f"[{self.name}]"

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:
        raise NotImplementedError  # pragma: no cover


class ListIndexVertex(_ListVertex):
    __slots__ = 'index'

    def __init__(self, parent, index: int):
        self.index = index
        super().__init__(parent, index)

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:
        data = parent_match.data
        if not isinstance(data, list):
            return None
        try:
            index = self.index
            value = data[index]
            return ListMatch(
                parent_match,
                index,
                value,
                self,
                vertex_index,
                parent_match.remembered_on_catch_match,
                parent_match.remembered_on_catch_action
            )
        except IndexError:
            pass

    @property
    def default_value_for_set(self) -> Union[dict, list]:
        return list()

    def set(self, parent_match: TraverserMatch, value: JsonTypes) -> TraverserMatch:
        data = parent_match.data
        if isinstance(data, list):
            try:
                data[self.name] = value
            except IndexError:
                if len(data) != self.name:
                    raise SetError(
                        self.parent,
                        f"The path {self} index is out of range.  "
                        f"The index must be in the range 0 >= index <={len(data)}.  To append the index must the list "
                        f"current length.",
                        self.path_segment
                    )
                data.append(value)
            return self.match(parent_match, None, parent_match.vertex_index + 1)
        else:
            self.raise_invalid_set(data, value)

    def pop(self, match: TraverserMatch) -> JsonTypes:
        data = match.parent.data
        if isinstance(data, list):
            return data.pop(self.name)
        else:
            self.raise_invalid_pop(data)


class ListSliceVertex(_ListVertex):
    __slots__ = '_slice'

    def __init__(self, parent, slice_: slice):
        self._slice = slice_
        super().__init__(parent, slice_)
        self.is_catch_vertex = True

    @property
    def path_segment(self):
        slice_ = self._slice
        start = slice_.start
        if not start:
            start = ''
        stop = slice_.stop
        if not stop:
            stop = ''
        step = slice_.step
        if not step:
            return f"[{start}:{stop}]"
        return f"[{start}:{stop}:{step}]"

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:

        remembered_catch_state = parent_match.remembered_catch_state

        # create a slice iterator  if there is none
        if not remembered_catch_state:
            data = parent_match.data
            if not isinstance(data, list):
                return None
            remembered_catch_state = enumerate_slice(self._slice, data)
            traverser.remember_on_catch(parent_match, remembered_catch_state)

        try:
            item = next(remembered_catch_state)
            return ListMatch(
                parent_match,
                item[0],
                item[1],
                self,
                vertex_index,
                parent_match.remembered_on_catch_match,
                parent_match.remembered_on_catch_action
            )
        except StopIteration:
            traverser.restore_on_catch(parent_match)
            return None


class ListWildVertex(_ListVertex):
    __slots__ = ()

    def __init__(self, parent):
        super().__init__(parent, '*')
        self.is_catch_vertex = True

    @property
    def path_segment(self):
        return "[*]"

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:

        remembered_catch_state = parent_match.remembered_catch_state

        # create a list iterator  if there is none
        if not remembered_catch_state:
            data = parent_match.data
            if not isinstance(data, list):
                return None
            remembered_catch_state = enumerate(data)
            traverser.remember_on_catch(parent_match, remembered_catch_state)

        try:
            item = next(remembered_catch_state)
            return ListMatch(
                parent_match,
                item[0],
                item[1],
                self,
                vertex_index,
                parent_match.remembered_on_catch_match,
                parent_match.remembered_on_catch_action
            )
        except StopIteration:
            traverser.restore_on_catch(parent_match)
            return None



from treepath.path.traverser.match_traverser import MatchTraverser
from treepath.path.typing.json_types import JsonTypes


class ValueTraverser(MatchTraverser):
    __slots__ = ()

    def __next__(self) -> JsonTypes:
        return super().__next__().data



from __future__ import annotations

from typing import Callable, Union, Any, Type, Optional, List, Generic

from treepath.descriptor.document import Document
from treepath.descriptor.document_iterator import DocumentIterator
from treepath.descriptor.document_list import DocumentList
from treepath.descriptor.path_descriptor import T, PathDescriptor
from treepath.path.builder.path_builder import PathBuilder, get_vertex_from_path_builder
from treepath.path.exceptions.set_error import SetError
from treepath.path.traverser.traverser_functions import get, set_, find, get_match, find_matches, set_match
from treepath.path.typing.json_arg_types import JsonArgTypes
from treepath.path.typing.json_types import JsonTypes
from treepath.path.utils.function import do_nothing


class DocTypeTransformer:
    __slots__ = '_type'

    def __init__(self, _type):
        self._type = _type

    def to_wrapped_value(self, json_: JsonArgTypes) -> Document:
        return self._type(json_)

    def to_json_value(self, doc: Document) -> JsonArgTypes:
        return doc.data


class DocTypeIteratorTransformer(Generic[T]):
    __slots__ = '_expression', '_to_wrapped_value'

    def __init__(self, expression, to_wrapped_value):
        self._expression = expression
        self._to_wrapped_value = to_wrapped_value

    def to_wrapped_value(self, inner_iterator) -> DocumentIterator[T]:
        outer_iterator = DocumentIterator[T](inner_iterator, to_wrapped_value=self._to_wrapped_value)
        return outer_iterator

    def to_json_value(self, doc: Document):
        vertex = get_vertex_from_path_builder(self._expression)
        raise SetError(
            vertex,
            f"The iterator descriptor for path '{self._expression}' does not support set",
            vertex.path_segment
        )


class DocTypeListTransformer(Generic[T]):
    __slots__ = '_to_wrapped_value', '_to_json_value'

    def __init__(self, to_wrapped_value, to_json_value):
        self._to_wrapped_value = to_wrapped_value
        self._to_json_value = to_json_value

    def to_wrapped_value(self, list_: List[JsonTypes]) -> DocumentList[T]:
        outer_iterator = DocumentList[T](
            list_,
            to_wrapped_value=self._to_wrapped_value,
            to_json_value=self._to_json_value
        )
        return outer_iterator

    def to_json_value(self, doc: DocumentList[T]):
        return doc.data


class DescriptorBuilder:
    """
    custom type = ct
    doc type  = dt
    itr = i
    single = s

    ct,s
    ct,i
    dt,s
    dt,i

    """

    def __init__(self,
                 type_: Type[T],
                 expression: Optional[PathBuilder] = None,
                 *,
                 getter: Union[get, find, get_match, find_matches] = get,
                 setter: Union[set_, set_match] = set_,
                 to_wrapped_value: Callable[[JsonTypes], Any] = do_nothing,
                 to_json_value: Callable[[Any], JsonTypes] = do_nothing,
                 is_for_json_list=False
                 ):
        self._type = type_
        self._expression = expression
        self._getter = getter
        self._setter = setter
        self._to_wrapped_value = to_wrapped_value
        self._to_json_value = to_json_value
        self._is_for_json_list = is_for_json_list
        self.__build = [
            [
                self.__build_doc_type_for_single,
                self.__build_doc_type_for_iterator,
                self.__build_doc_type_for_list
            ],
            [
                self.__custom_type_for_single,
                self.__custom_type_for_iterator,
                self.__custom_type_for_list,
            ],
        ]

    def build_single(self) -> PathDescriptor:
        type_index = self.__index_for_type_column()
        return self.__build[type_index][0]()

    def build_iterator(self) -> PathDescriptor:
        type_index = self.__index_for_type_column()
        return self.__build[type_index][1]()

    def build_list(self) -> PathDescriptor:
        type_index = self.__index_for_type_column()
        return self.__build[type_index][2]()

    def __index_for_type_column(self) -> int:
        if issubclass(self._type, Document):
            return 0
        else:
            return 1

    def __build_path_descriptor(self):
        return PathDescriptor(
            expression=self._expression,
            getter=self._getter,
            setter=self._setter,
            to_wrapped_value=self._to_wrapped_value,
            to_json_value=self._to_json_value
        )

    def __build_doc_type_for_single(self) -> PathDescriptor:
        doc_type_transformer = DocTypeTransformer(self._type)
        self._to_wrapped_value = doc_type_transformer.to_wrapped_value
        self._to_json_value = doc_type_transformer.to_json_value
        return self.__build_path_descriptor()

    def __build_doc_type_for_iterator(self) -> PathDescriptor:
        doc_type_transformer = DocTypeTransformer(self._type)
        doc_list_type_transformer = DocTypeIteratorTransformer(self._expression, doc_type_transformer.to_wrapped_value)
        self._to_wrapped_value = doc_list_type_transformer.to_wrapped_value
        self._to_json_value = doc_list_type_transformer.to_json_value
        return self.__build_path_descriptor()

    def __build_doc_type_for_list(self) -> PathDescriptor:
        doc_type_transformer = DocTypeTransformer(self._type)
        doc_list_type_transformer = DocTypeListTransformer(
            doc_type_transformer.to_wrapped_value,
            doc_type_transformer.to_json_value
        )
        self._to_wrapped_value = doc_list_type_transformer.to_wrapped_value
        self._to_json_value = doc_list_type_transformer.to_json_value
        return self.__build_path_descriptor()

    def __custom_type_for_single(self) -> PathDescriptor:
        return self.__build_path_descriptor()

    def __custom_type_for_iterator(self) -> PathDescriptor:
        doc_list_type_transformer = DocTypeIteratorTransformer(self._expression, self._to_wrapped_value)
        self._to_wrapped_value = doc_list_type_transformer.to_wrapped_value
        self._to_json_value = doc_list_type_transformer.to_json_value

        return self.__build_path_descriptor()

    def __custom_type_for_list(self) -> PathDescriptor:
        doc_list_type_transformer = DocTypeListTransformer(
            self._to_wrapped_value,
            self._to_json_value
        )
        self._to_wrapped_value = doc_list_type_transformer.to_wrapped_value
        self._to_json_value = doc_list_type_transformer.to_json_value
        return self.__build_path_descriptor()



from __future__ import annotations

from typing import cast, overload, Type, Generic, Callable, TypeVar, Union, Optional

from treepath.descriptor.abstract_document import AbstractDocument
from treepath.path.builder.path_builder import PathBuilder
from treepath.path.builder.root_path_builder import RootPathBuilder
from treepath.path.traverser.traverser_functions import get, set_, pop, find, get_match, find_matches, set_match
from treepath.path.typing.json_types import JsonTypes
from treepath.path.utils.function import do_nothing

T = TypeVar('T')


class PathDescriptor(Generic[T]):
    """
    Treepath implementation of the python descriptor.
    See details on descriptors: https://docs.python.org/3/howto/descriptor.html

    See descriptor_functions for usage.
    """
    __slots__ = "_expression", "_getter", "_setter", "_to_wrapped_value", "_to_json_value"

    def __init__(self,
                 expression: Optional[PathBuilder] = None,
                 *,
                 getter: Union[get, find, get_match, find_matches] = get,
                 setter: Union[set_, set_match] = set_,
                 to_wrapped_value: Callable[[JsonTypes], T] = do_nothing,
                 to_json_value: Callable[[T], JsonTypes] = do_nothing
                 ):
        self._expression = expression
        self._getter = getter
        self._setter = setter
        self._to_wrapped_value = to_wrapped_value
        self._to_json_value = to_json_value

    def __set_name__(self, owner: Type[AbstractDocument], name: str) -> None:
        if not issubclass(owner, AbstractDocument):
            raise ValueError(
                f"{type(self)} can only be assign to a class that implements "
                f"{AbstractDocument}")
        if self._expression is None:
            self._expression = RootPathBuilder()[name]

    @overload
    def __get__(self, owner_obj: None, owner_obj_type: None) -> PathDescriptor:  # pragma: no cover
        ...

    @overload
    def __get__(self, owner_obj: AbstractDocument, owner_obj_type: Type[AbstractDocument]) -> T:  # pragma: no cover
        ...

    def __get__(
            self, owner_obj: AbstractDocument | None, owner_obj_type: Type[AbstractDocument] | None = None
    ) -> PathDescriptor | T:
        if owner_obj is None:
            return self
        json_value = self._getter(self._expression, owner_obj.data)
        wrapped_value = self._to_wrapped_value(json_value)
        return cast(T, wrapped_value)

    def __set__(self, owner_obj: AbstractDocument, wrapped_value: T) -> None:
        json_value = self._to_json_value(wrapped_value)
        self._setter(self._expression, json_value, owner_obj.data)

    def __delete__(self, owner_obj: AbstractDocument) -> None:
        pop(self._expression, owner_obj.data)



import json

from treepath.descriptor.abstract_document import AbstractDocument
from treepath.path.typing.json_arg_types import JsonArgTypes


class Document(AbstractDocument):
    """
    A class that wraps a json document and provides common json operation.
    """
    __slots__ = "_data"

    def __init__(self, data: JsonArgTypes):
        self._data = data

    @property
    def data(self) -> JsonArgTypes:
        return self._data

    @data.setter
    def data(self, data: JsonArgTypes):
        self._data = data

    @property
    def json_str(self) -> str:
        return json.dumps(self.data)

    @property
    def pretty_json_str(self) -> str:
        return json.dumps(self.data, sort_keys=True, indent=2)

    def __repr__(self):
        return self.json_str

    def __str__(self):
        return f"{self.__class__.__name__}:" \
               f" json: {self.pretty_json_str}"

    def __eq__(self, other):
        return isinstance(other, AbstractDocument) and self.data == other.data

    def __ne__(self, other):
        return (not isinstance(other, AbstractDocument)) or self.data != other.data
